@@ -26,12 +26,30 @@
 #include "config.h"
 
 
+/*
+ * Structure for network configuration parsing. This data is used to implement
+ * a generic parser for each network block variable. The table of configuration
+ * variables is defined below in this file (ssid_fields[]).
+ */
 struct parse_data {
+	/* Configuration variable name */
 	char *name;
+
+	/* Parser function for this variable */
 	int (*parser)(const struct parse_data *data, struct wpa_ssid *ssid,
 		      int line, const char *value);
+
+	/* Writer function (i.e., to get the variable in text format from
+	 * internal presentation). */
 	char * (*writer)(const struct parse_data *data, struct wpa_ssid *ssid);
+
+	/* Variable specific parameters for the parser. */
 	void *param1, *param2, *param3, *param4;
+
+	/* 0 = this variable can be included in debug output
+	 * 1 = this variable contains key/private data and it must not be
+	 *     included in debug output unless explicitly requested
+	 */
 	int key_data;
 };
 
@@ -41,7 +59,7 @@ static char * wpa_config_parse_string(const char *value, size_t *len)
 	if (*value == '"') {
 		char *pos;
 		value++;
-		pos = strchr(value, '"');
+		pos = strrchr(value, '"');
 		if (pos == NULL || pos[1] != '\0')
 			return NULL;
 		*pos = '\0';
@@ -79,7 +97,8 @@ static int wpa_config_parse_str(const struct parse_data *data,
 	*dst = wpa_config_parse_string(value, &res_len);
 	if (*dst == NULL) {
 		wpa_printf(MSG_ERROR, "Line %d: failed to parse %s '%s'.",
-			   line, data->name, value);
+			   line, data->name,
+			   data->key_data ? "[KEY DATA REMOVED]" : value);
 		return -1;
 	}
 	if (data->param2)
@@ -275,14 +294,15 @@ static int wpa_config_parse_psk(const struct parse_data *data,
 				const char *value)
 {
 	if (*value == '"') {
-		char *pos;
+		const char *pos;
 		size_t len;
 
 		value++;
 		pos = strrchr(value, '"');
 		if (pos)
-			*pos = '\0';
-		len = strlen(value);
+			len = pos - value;
+		else
+			len = strlen(value);
 		if (len < 8 || len > 63) {
 			wpa_printf(MSG_ERROR, "Line %d: Invalid passphrase "
 				   "length %lu (expected: 8..63) '%s'.",
@@ -291,8 +311,12 @@ static int wpa_config_parse_psk(const struct parse_data *data,
 		}
 		wpa_hexdump_ascii_key(MSG_MSGDUMP, "PSK (ASCII passphrase)",
 				      (u8 *) value, len);
-		ssid->passphrase = strdup(value);
-		return ssid->passphrase == NULL ? -1 : 0;
+		ssid->passphrase = malloc(len + 1);
+		if (ssid->passphrase == NULL)
+			return -1;
+		memcpy(ssid->passphrase, value, len);
+		ssid->passphrase[len] = '\0';
+		return 0;
 	}
 
 	if (hexstr2bin(value, ssid->psk, PMK_LEN) ||
@@ -756,6 +780,7 @@ static int wpa_config_parse_eap(const struct parse_data *data,
 		methods = realloc(methods, num_methods + 1);
 		if (methods == NULL) {
 			free(tmp);
+			free(buf);
 			return -1;
 		}
 		methods[num_methods] = eap_get_type(start);
@@ -925,20 +950,63 @@ static char * wpa_config_write_wep_key3(const struct parse_data *data,
 }
 
 
+/* Helper macros for network block parser */
+
+/* OFFSET: Get offset of a variable within the wpa_ssid structure */
 #define OFFSET(v) ((void *) &((struct wpa_ssid *) 0)->v)
+
+/* STR: Define a string variable for an ASCII string; f = field name */
 #define STR(f) .name = #f, .parser = wpa_config_parse_str, \
 	.writer = wpa_config_write_str, .param1 = OFFSET(f)
+
+/* STR_LEN: Define a string variable with a separate variable for storing the
+ * data length. Unlike STR(), this can be used to store arbitrary binary data
+ * (i.e., even nul termination character). */
 #define STR_LEN(f) STR(f), .param2 = OFFSET(f ## _len)
+
+/* STR_RANGE: Like STR_LEN(), but with minimum and maximum allowed length
+ * explicitly specified. */
 #define STR_RANGE(f, min, max) STR_LEN(f), .param3 = (void *) (min), \
 	.param4 = (void *) (max)
+
+
+/* INT: Define an integer variable */
 #define INT(f) .name = #f, .parser = wpa_config_parse_int, \
 	.writer = wpa_config_write_int, \
 	.param1 = OFFSET(f), .param2 = (void *) 0
+
+/* INT: Define an integer variable with allowed value range */
 #define INT_RANGE(f, min, max) INT(f), .param3 = (void *) (min), \
 	.param4 = (void *) (max)
+
+/* FUNC: Define a configuration variable that uses a custom function for
+ * parsing and writing the value. */
 #define FUNC(f) .name = #f, .parser = wpa_config_parse_ ## f, \
 		.writer = wpa_config_write_ ## f
 
+/*
+ * Table of network configuration variables. This table is used to parse each
+ * network configuration variable, e.g., each line in wpa_supplicant.conf file
+ * that is inside a network block.
+ *
+ * This table is generated using the helper macros defined above and with
+ * generous help from the C pre-processor. The field name is stored as a string
+ * into .name and for STR and INT types, the offset of the target buffer within
+ * struct wpa_ssid is stored in .param1. .param2 (if not NULL) is similar
+ * offset to the field containing the length of the configuration variable.
+ * .param3 and .param4 can be used to mark the allowed range (length for STR
+ * and value for INT).
+ *
+ * For each configuration line in wpa_supplicant.conf, the parser goes through
+ * this table and select the entry that matches with the field name. The parser
+ * function (.parser) is then called to parse the actual value of the field.
+ *
+ * This kind of mechanism makes it easy to add new configuration parameters,
+ * since only one line needs to be added into this table and into the
+ * struct wpa_ssid definition if the new variable is either a string or
+ * integer. More complex types will need to use their own parser and writer
+ * functions.
+ */
 static const struct parse_data ssid_fields[] = {
 	{ STR_RANGE(ssid, 0, MAX_SSID_LEN) },
 	{ INT_RANGE(scan_ssid, 0, 1) },
@@ -1005,6 +1073,7 @@ static const struct parse_data ssid_fields[] = {
 /**
  * wpa_config_add_prio_network - Add a network to priority lists
  * @config: Configuration data from wpa_config_read()
+ * @ssid: Pointer to the network configuration to be added to the list
  * Returns: 0 on success, -1 on failure
  *
  * This function is used to add a network block to the priority list of
@@ -1018,6 +1087,10 @@ int wpa_config_add_prio_network(struct wpa_config *config,
 	int prio;
 	struct wpa_ssid *prev, **nlist;
 
+	/*
+	 * Add to an existing priority list if one is available for the
+	 * configured priority level for this network.
+	 */
 	for (prio = 0; prio < config->num_prio; prio++) {
 		prev = config->pssid[prio];
 		if (prev->priority == ssid->priority) {
@@ -1028,7 +1101,7 @@ int wpa_config_add_prio_network(struct wpa_config *config,
 		}
 	}
 
-	/* First network for this priority - add new priority list */
+	/* First network for this priority - add a new priority list */
 	nlist = realloc(config->pssid,
 			(config->num_prio + 1) * sizeof(struct wpa_ssid *));
 	if (nlist == NULL)
@@ -1084,7 +1157,7 @@ static int wpa_config_update_prio_list(struct wpa_config *config)
  * wpa_config_free_ssid - Free network/ssid configuration data
  * @ssid: Configuration data for the network
  *
- * This function frees all resources allocated for the netowkr configuration
+ * This function frees all resources allocated for the network configuration
  * data.
  */
 void wpa_config_free_ssid(struct wpa_ssid *ssid)
@@ -1165,7 +1238,7 @@ void wpa_config_free(struct wpa_config *config)
 
 /**
  * wpa_config_allowed_eap_method - Check whether EAP method is allowed
- * @ssid: Pointer to a configuration data
+ * @ssid: Pointer to configuration data
  * @method: EAP type
  * Returns: 1 = allowed EAP method, 0 = not allowed
  */
@@ -1277,7 +1350,7 @@ int wpa_config_remove_network(struct wpa_config *config, int id)
 
 /**
  * wpa_config_set_network_defaults - Set network default values
- * @ssid: Pointer to a network configuration data
+ * @ssid: Pointer to network configuration data
  */
 void wpa_config_set_network_defaults(struct wpa_ssid *ssid)
 {
@@ -1292,7 +1365,7 @@ void wpa_config_set_network_defaults(struct wpa_ssid *ssid)
 
 /**
  * wpa_config_set - Set a variable in network configuration
- * @ssid: Pointer to a network configuration data
+ * @ssid: Pointer to network configuration data
  * @var: Variable name, e.g., "ssid"
  * @value: Variable value
  * @line: Line number in configuration file or 0 if not used
@@ -1339,7 +1412,7 @@ int wpa_config_set(struct wpa_ssid *ssid, const char *var, const char *value,
 
 /**
  * wpa_config_get - Get a variable in network configuration
- * @ssid: Pointer to a network configuration data
+ * @ssid: Pointer to network configuration data
  * @var: Variable name, e.g., "ssid"
  * Returns: Value of the variable or %NULL on failure
  *
@@ -1368,7 +1441,7 @@ char * wpa_config_get(struct wpa_ssid *ssid, const char *var)
 
 /**
  * wpa_config_update_psk - Update WPA PSK based on passphrase and SSID
- * @ssid: Pointer to a network configuration data
+ * @ssid: Pointer to network configuration data
  *
  * This function must be called to update WPA PSK when either SSID or the
  * passphrase has changed for the network configuration.
