@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - driver interaction with generic Linux Wireless Extensions
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -51,6 +51,10 @@ struct wpa_driver_wext_data {
 	struct wpa_driver_capa capa;
 	int has_capability;
 	int we_version_compiled;
+
+	/* for set_auth_alg fallback */
+	int use_crypt;
+	int auth_alg_fallback;
 };
 
 
@@ -73,7 +77,7 @@ static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 		perror("ioctl[SIOCSIWAUTH]");
 		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
 			idx, value);
-		ret = -1;
+		ret = errno == EOPNOTSUPP ? -2 : -1;
 	}
 
 	return ret;
@@ -154,8 +158,18 @@ int wpa_driver_wext_get_ssid(void *priv, u8 *ssid)
 	if (ioctl(drv->ioctl_sock, SIOCGIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCGIWESSID]");
 		ret = -1;
-	} else
+	} else {
 		ret = iwr.u.essid.length;
+		if (ret > 32)
+			ret = 32;
+		/* Some drivers include nul termination in the SSID, so let's
+		 * remove it here before further processing. WE-21 changes this
+		 * to explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ret > 0 && ssid[ret - 1] == '\0' &&
+		    drv->we_version_compiled < 21)
+			ret--;
+	}
 
 	return ret;
 }
@@ -185,12 +199,19 @@ int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 	memset(buf, 0, sizeof(buf));
 	memcpy(buf, ssid, ssid_len);
 	iwr.u.essid.pointer = (caddr_t) buf;
-	/* For historic reasons, set SSID length to include one extra
-	 * character, C string nul termination, even though SSID is really an
-	 * octet string that should not be presented as a C string. Some Linux
-	 * drivers decrement the length by one and can thus end up missing the
-	 * last octet of the SSID if the length is not incremented here. */
-	iwr.u.essid.length = ssid_len ? ssid_len + 1 : 0;
+	if (drv->we_version_compiled < 21) {
+		/* For historic reasons, set SSID length to include one extra
+		 * character, C string nul termination, even though SSID is
+		 * really an octet string that should not be presented as a C
+		 * string. Some Linux drivers decrement the length by one and
+		 * can thus end up missing the last octet of the SSID if the
+		 * length is not incremented here. WE-21 changes this to
+		 * explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ssid_len)
+			ssid_len++;
+	}
+	iwr.u.essid.length = ssid_len;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCSIWESSID]");
@@ -911,7 +932,11 @@ static int wpa_scan_result_compar(const void *a, const void *b)
 
 	/* use freq for channel preference */
 
-	/* all things being equal, use signal level */
+	/* all things being equal, use signal level; if signal levels are
+	 * identical, use quality values since some drivers may only report
+	 * that value and leave the signal level zero */
+	if (wb->level == wa->level)
+		return wb->qual - wb->qual;
 	return wb->level - wa->level;
 }
 
@@ -1443,6 +1468,7 @@ static int wpa_driver_wext_set_drop_unencrypted(void *priv,
 {
 	struct wpa_driver_wext_data *drv = priv;
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
+	drv->use_crypt = enabled;
 	return wpa_driver_wext_set_auth_param(drv, IW_AUTH_DROP_UNENCRYPTED,
 					      enabled);
 }
@@ -1546,6 +1572,50 @@ static int wpa_driver_wext_keymgmt2wext(int keymgmt)
 }
 
 
+static int
+wpa_driver_wext_auth_alg_fallback(struct wpa_driver_wext_data *drv,
+				  struct wpa_driver_associate_params *params)
+{
+	struct iwreq iwr;
+	int ret = 0;
+
+	wpa_printf(MSG_DEBUG, "WEXT: Driver did not support "
+		   "SIOCSIWAUTH for AUTH_ALG, trying SIOCSIWENCODE");
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
+	/* Just changing mode, not actual keys */
+	iwr.u.encoding.flags = 0;
+	iwr.u.encoding.pointer = (caddr_t) NULL;
+	iwr.u.encoding.length = 0;
+
+	/*
+	 * Note: IW_ENCODE_{OPEN,RESTRICTED} can be interpreted to mean two
+	 * different things. Here they are used to indicate Open System vs.
+	 * Shared Key authentication algorithm. However, some drivers may use
+	 * them to select between open/restricted WEP encrypted (open = allow
+	 * both unencrypted and encrypted frames; restricted = only allow
+	 * encrypted frames).
+	 */
+
+	if (!drv->use_crypt) {
+		iwr.u.encoding.flags |= IW_ENCODE_DISABLED;
+	} else {
+		if (params->auth_alg & AUTH_ALG_OPEN_SYSTEM)
+			iwr.u.encoding.flags |= IW_ENCODE_OPEN;
+		if (params->auth_alg & AUTH_ALG_SHARED_KEY)
+			iwr.u.encoding.flags |= IW_ENCODE_RESTRICTED;
+	}
+
+	if (ioctl(drv->ioctl_sock, SIOCSIWENCODE, &iwr) < 0) {
+		perror("ioctl[SIOCSIWENCODE]");
+		ret = -1;
+	}
+
+	return ret;
+}
+
+
 static int
 wpa_driver_wext_associate(void *priv,
 			  struct wpa_driver_associate_params *params)
@@ -1557,6 +1627,14 @@ wpa_driver_wext_associate(void *priv,
 
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
 
+	/*
+	 * If the driver did not support SIOCSIWAUTH, fallback to
+	 * SIOCSIWENCODE here.
+	 */
+	if (drv->auth_alg_fallback &&
+	    wpa_driver_wext_auth_alg_fallback(drv, params) < 0)
+		ret = -1;
+
 	if (!params->bssid &&
 	    wpa_driver_wext_set_bssid(drv, NULL) < 0)
 		ret = -1;
@@ -1590,7 +1668,10 @@ wpa_driver_wext_associate(void *priv,
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_KEY_MGMT, value) < 0)
 		ret = -1;
-	value = params->key_mgmt_suite != KEY_MGMT_NONE;
+	value = params->key_mgmt_suite != KEY_MGMT_NONE ||
+		params->pairwise_suite != CIPHER_NONE ||
+		params->group_suite != CIPHER_NONE ||
+		params->wpa_ie_len;
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_PRIVACY_INVOKED, value) < 0)
 		ret = -1;
@@ -1623,7 +1704,7 @@ wpa_driver_wext_associate(void *priv,
 static int wpa_driver_wext_set_auth_alg(void *priv, int auth_alg)
 {
 	struct wpa_driver_wext_data *drv = priv;
-	int algs = 0;
+	int algs = 0, res;
 
 	if (auth_alg & AUTH_ALG_OPEN_SYSTEM)
 		algs |= IW_AUTH_ALG_OPEN_SYSTEM;
@@ -1636,8 +1717,10 @@ static int wpa_driver_wext_set_auth_alg(void *priv, int auth_alg)
 		algs = IW_AUTH_ALG_OPEN_SYSTEM;
 	}
 
-	return wpa_driver_wext_set_auth_param(drv,
-					      IW_AUTH_80211_AUTH_ALG, algs);
+	res = wpa_driver_wext_set_auth_param(drv, IW_AUTH_80211_AUTH_ALG,
+					     algs);
+	drv->auth_alg_fallback = res == -2;
+	return res;
 }
 
 
