@@ -1,3 +1,17 @@
+/*
+ * WPA Supplicant - driver_wext exported functions
+ * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2 as
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
+ */
+
 #ifndef DRIVER_WEXT_H
 #define DRIVER_WEXT_H
 
@@ -22,6 +36,9 @@ int wpa_driver_wext_get_scan_results(void *priv,
 
 void wpa_driver_wext_scan_timeout(void *eloop_ctx, void *timeout_ctx);
 
+int wpa_driver_wext_alternative_ifindex(struct wpa_driver_wext_data *drv,
+					const char *ifname);
+
 void * wpa_driver_wext_init(void *ctx, const char *ifname);
 void wpa_driver_wext_deinit(void *priv);
 
