@@ -24,10 +24,6 @@
 #include "eap_sim_common.h"
 
 
-#define MSK_LEN 8
-#define EMSK_LEN 8
-
-
 static void eap_sim_prf(const u8 *key, u8 *x, size_t xlen)
 {
 	u8 xkey[64];
@@ -87,18 +83,13 @@ void eap_sim_derive_keys(const u8 *mk, u8 *k_encr, u8 *k_aut, u8 *msk)
 	memcpy(k_aut, pos, EAP_SIM_K_AUT_LEN);
 	pos += EAP_SIM_K_AUT_LEN;
 	memcpy(msk, pos, EAP_SIM_KEYING_DATA_LEN);
-	pos += MSK_LEN;
 
 	wpa_hexdump_key(MSG_DEBUG, "EAP-SIM: K_encr",
 			k_encr, EAP_SIM_K_ENCR_LEN);
 	wpa_hexdump_key(MSG_DEBUG, "EAP-SIM: K_aut",
 			k_aut, EAP_SIM_K_ENCR_LEN);
-	wpa_hexdump_key(MSG_DEBUG, "EAP-SIM: MSK",
-			msk, MSK_LEN);
-	wpa_hexdump_key(MSG_DEBUG, "EAP-SIM: Ext. MSK",
-			msk + MSK_LEN, EMSK_LEN);
 	wpa_hexdump_key(MSG_DEBUG, "EAP-SIM: keying material",
-		    msk, EAP_SIM_KEYING_DATA_LEN);
+			msk, EAP_SIM_KEYING_DATA_LEN);
 }
 
 
@@ -135,8 +126,6 @@ void eap_sim_derive_keys_reauth(u16 _counter,
 	wpa_hexdump(MSG_DEBUG, "EAP-SIM: XKEY'", xkey, SHA1_MAC_LEN);
 
 	eap_sim_prf(xkey, msk, EAP_SIM_KEYING_DATA_LEN);
-	wpa_hexdump(MSG_DEBUG, "EAP-SIM: MSK", msk, MSK_LEN);
-	wpa_hexdump(MSG_DEBUG, "EAP-SIM: Ext. MSK", msk + MSK_LEN, EMSK_LEN);
 	wpa_hexdump(MSG_DEBUG, "EAP-SIM: keying material",
 		    msk, EAP_SIM_KEYING_DATA_LEN);
 }
@@ -492,6 +481,7 @@ u8 * eap_sim_parse_encr(const u8 *k_encr, const u8 *encr_data,
 	decrypted = malloc(encr_data_len);
 	if (decrypted == NULL)
 		return NULL;
+	memcpy(decrypted, encr_data, encr_data_len);
 
 	aes_128_cbc_decrypt(k_encr, iv, decrypted, encr_data_len);
 	wpa_hexdump(MSG_MSGDUMP, "EAP-SIM: Decrypted AT_ENCR_DATA",
