@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - test code
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -82,7 +82,7 @@ static struct eapol_test_data eapol_test;
 static void send_eap_request_identity(void *eloop_ctx, void *timeout_ctx);
 
 
-void hostapd_logger(void *ctx, u8 *addr, unsigned int module, int level,
+void hostapd_logger(void *ctx, const u8 *addr, unsigned int module, int level,
 		    char *fmt, ...)
 {
 	char *format;
@@ -651,25 +651,6 @@ ieee802_1x_receive_auth(struct radius_msg *msg, struct radius_msg *req,
 }
 
 
-static int wpa_supplicant_scard_init(struct wpa_supplicant *wpa_s,
-				     struct wpa_ssid *ssid)
-{
-	if (ssid->pcsc == NULL || wpa_s->scard != NULL)
-		return 0;
-	wpa_printf(MSG_DEBUG, "Selected network is configured to use SIM - "
-		   "initialize PCSC");
-	wpa_s->scard = scard_init(SCARD_TRY_BOTH);
-	if (wpa_s->scard == NULL) {
-		wpa_printf(MSG_WARNING, "Failed to initialize SIM "
-			   "(pcsc-lite)");
-		return -1;
-	}
-	eapol_sm_register_scard_ctx(wpa_s->eapol, wpa_s->scard);
-
-	return 0;
-}
-
-
 static void wpa_init_conf(struct eapol_test_data *e,
 			  struct wpa_supplicant *wpa_s, const char *authsrv,
 			  int port, const char *secret)
