@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - RSN pre-authentication and PMKSA caching
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -74,7 +74,7 @@ static void pmksa_cache_set_expiration(struct wpa_sm *sm);
 
 
 static void pmksa_cache_free_entry(struct wpa_sm *sm,
-				   struct rsn_pmksa_cache *entry)
+				   struct rsn_pmksa_cache *entry, int replace)
 {
 	int current;
 
@@ -88,6 +88,15 @@ static void pmksa_cache_free_entry(struct wpa_sm *sm,
 	if (current) {
 		wpa_printf(MSG_DEBUG, "RSN: removed current PMKSA entry");
 		sm->cur_pmksa = NULL;
+
+		if (replace) {
+			/* A new entry is being added, so no need to
+			 * deauthenticate in this case. This happens when EAP
+			 * authentication is completed again (reauth or failed
+			 * PMKSA caching attempt). */
+			return;
+		}
+
 		memset(sm->pmk, 0, sizeof(sm->pmk));
 		wpa_sm_deauthenticate(sm, REASON_UNSPECIFIED);
 		wpa_sm_req_scan(sm, 0, 0);
@@ -106,7 +115,7 @@ static void pmksa_cache_expire(void *eloop_ctx, void *timeout_ctx)
 		sm->pmksa = entry->next;
 		wpa_printf(MSG_DEBUG, "RSN: expired PMKSA cache entry for "
 			   MACSTR, MAC2STR(entry->aa));
-		pmksa_cache_free_entry(sm, entry);
+		pmksa_cache_free_entry(sm, entry, 0);
 	}
 
 	pmksa_cache_set_expiration(sm);
@@ -205,7 +214,9 @@ pmksa_cache_add(struct wpa_sm *sm, const u8 *pmk,
 				sm->pmksa = pos->next;
 			else
 				prev->next = pos->next;
-			pmksa_cache_free_entry(sm, pos);
+			wpa_printf(MSG_DEBUG, "RSN: Replace PMKSA entry for "
+				   "the current AP");
+			pmksa_cache_free_entry(sm, pos, 1);
 			break;
 		}
 		prev = pos;
@@ -220,7 +231,7 @@ pmksa_cache_add(struct wpa_sm *sm, const u8 *pmk,
 			   "entry (for " MACSTR ") to make room for new one",
 			   MAC2STR(pos->aa));
 		wpa_sm_remove_pmkid(sm, pos->aa, pos->pmkid);
-		pmksa_cache_free_entry(sm, pos);
+		pmksa_cache_free_entry(sm, pos, 0);
 	}
 
 	/* Add the new entry; order by expiration time */
@@ -393,7 +404,7 @@ void pmksa_cache_clear_current(struct wpa_sm *sm)
  * pmksa_cache_set_current - Set the current PMKSA entry selection
  * @sm: Pointer to WPA state machine data from wpa_sm_init()
  * @pmkid: PMKID for selecting PMKSA or %NULL if not used
- * @bssid: BSSID for PMKSA  or %NULL if not used
+ * @bssid: BSSID for PMKSA or %NULL if not used
  * @ssid: The network configuration for the current network
  * @try_opportunistic: Whether to allow opportunistic PMKSA caching
  * Returns: 0 if PMKSA was found or -1 if no matching entry was found
@@ -407,7 +418,7 @@ int pmksa_cache_set_current(struct wpa_sm *sm, const u8 *pmkid,
 		sm->cur_pmksa = pmksa_cache_get(sm, NULL, pmkid);
 	if (sm->cur_pmksa == NULL && bssid)
 		sm->cur_pmksa = pmksa_cache_get(sm, bssid, NULL);
-	if (sm->cur_pmksa == NULL && try_opportunistic)
+	if (sm->cur_pmksa == NULL && try_opportunistic && bssid)
 		sm->cur_pmksa = pmksa_cache_get_opportunistic(sm, ssid, bssid);
 	if (sm->cur_pmksa) {
 		wpa_hexdump(MSG_DEBUG, "RSN: PMKID",
